from athina.interfaces.model import Model
from ..ragas_evaluator import RagasEvaluator
from athina.evals.eval_type import RagasEvalTypeId
from athina.metrics.metric_type import MetricType
from ragas.metrics import context_recall
from typing import List, Optional

"""
RAGAS Context Recall Docs: https://docs.ragas.io/en/latest/concepts/metrics/context_recall.html
RAGAS Context Recall Github: https://github.com/explodinggradients/ragas/blob/main/src/ragas/metrics/_context_recall.py
"""


class RagasContextRecall(RagasEvaluator):
    """
    This measures the extent to which the retrieved context aligns with the annotated answer, treated as the ground truth.
    """

    @property
    def name(self):
        return RagasEvalTypeId.RAGAS_CONTEXT_RECALL.value

    @property
    def display_name(self):
        return "Ragas Context Recall"

    @property
    def metric_ids(self) -> List[str]:
        return [MetricType.RAGAS_CONTEXT_RECALL.value]

    @property
    def ragas_metric(self):
        return context_recall

    @property
    def ragas_metric_name(self):
        return "context_recall"

    @property
    def default_model(self):
        return Model.GPT35_TURBO.value

    @property
    def required_args(self):
        return ["query", "context", "expected_response"]

    @property
    def examples(self):
        return None

    @property
    def grade_reason(self) -> str:
        return "Context Recall metric is calculated by dividing the number of sentences in the ground truth that can be attributed to retrieved context by the total number of sentences in the grouund truth"

    def is_failure(self, score) -> Optional[bool]:
        return (
            bool(score < self._failure_threshold)
            if self._failure_threshold is not None
            else None
        )

    def generate_data_to_evaluate(
        self, context, query, expected_response, **kwargs
    ) -> dict:
        """
        Generates data for evaluation.

        :param context: list of strings of retrieved context
        :param query: user query
        :param expected_response: expected output
        :return: A dictionary with formatted data for evaluation
        """
        data = {
            "contexts": [context],
            "question": [query],
            "ground_truths": [[expected_response]],
        }
        return data



# Step to make an external api call
import json
import time
from typing import Union, Dict, List, Any, Iterable, Optional
import requests
from athina.steps import Step
from jinja2 import Environment
from athina.helpers.jinja_helper import PreserveUndefined


def prepare_input_data(data):
    return {key: json.dumps(value) if isinstance(value, (list, dict)) else value
        for key, value in data.items()}


class SpiderCrawl(Step):
    """
    Step that makes a crawl API Call to https://api.spider.cloud/crawl.

    Attributes:
        url: The query string.
        limit: The maximum amount of pages allowed to crawl per website. Remove the value or set it to 0 to crawl all pages. Defaults to 0.
        metadata: Collect metadata about the content found like page title, description, keywards and etc. This could help improve AI interoperability. Defaults to false.
        return_format: The format of the response. Defaults to raw.
        spider_key: The API key to use for the request.
    """

    url: str
    type: Optional[str] = "neural"
    limit: Optional[int] = 1
    metadata: Optional[bool] = False
    return_format: Optional[str] = "markdown"
    spider_key: str
    env: Environment = None

    class Config:
        arbitrary_types_allowed = True
       

    def execute(self, input_data: Any) -> Union[Dict[str, Any], None]:
        """Make an Search API call and return the response."""

        if input_data is None:
            input_data = {}

        if not isinstance(input_data, dict):
            raise TypeError("Input data must be a dictionary.")

        # Create a custom Jinja2 environment with double curly brace delimiters and PreserveUndefined
        self.env = Environment(
            variable_start_string='{{', 
            variable_end_string='}}',
            undefined=PreserveUndefined
        )

        body={
            "url": self.url,
            "limit": self.limit,
            "metadata": self.metadata,
            "return_format": self.return_format
        }
        prepared_body = None
        # Add a filter to the Jinja2 environment to convert the input data to JSON
        body_template = self.env.from_string(json.dumps(body))
        prepared_input_data = prepare_input_data(input_data)
        prepared_body = body_template.render(**prepared_input_data)
        

        retries = 2  # number of retries
        timeout = 300  # seconds
        for attempt in range(retries):
            try:
                response = requests.post(
                    url= "https://api.spider.cloud/crawl",
                    headers={
                        "Content-Type": "application/json",
                      'Authorization': f'Bearer {self.spider_key}',
                    },
                    json=json.loads(prepared_body, strict=False) if prepared_body else None,
                    timeout=timeout,
                )
                if response.status_code >= 400:
                    # If the status code is an error, return the error message
                    return {
                        "status": "error",
                        "data": f"Failed to make the API call.\nStatus code: {response.status_code}\nError:\n{response.text}",
                    }
                try:
                    json_response = response.json()
                    # If the response is JSON, return the JSON data

                    # Loop through the json response and get the content
                    content = []
                    for item in json_response:
                        value = {
                            "content": item.get('content'),
                            "url": item.get('url'),
                            "error": item.get('error'),
                        }
                        content.append(value)

                    
                    return {
                        "status": "success",
                        "data": content,
                    }
                
                except json.JSONDecodeError:
                    # If the response is not JSON, return the text
                    return {
                        "status": "success",
                        "data": response.text,
                    }
            except requests.Timeout:
                if attempt < retries - 1:
                    time.sleep(2)
                    continue
                # If the request times out after multiple attempts, return an error message
                return {
                    "status": "error",
                    "data": "Failed to make the API call.\nRequest timed out after multiple attempts.",
                }
            except Exception as e:
                # If an exception occurs, return the error message
                return {
                    "status": "error",
                    "data": f"Failed to make the API call.\nError: {e.__class__.__name__}\nDetails:\n{str(e)}",
                }



from athina.steps.base import Step, Fn, Debug
from athina.steps.conditional import Assert, If
from athina.steps.chain import Chain
from athina.steps.iterator import Map
from athina.steps.llm import PromptExecution
from athina.steps.api import ApiCall
from athina.steps.extract_entities import ExtractEntities
from athina.steps.classify_text import ClassifyText
from athina.steps.pinecone_retrieval import PineconeRetrieval
from athina.steps.qdrant_retrieval import QdrantRetrieval
from athina.steps.weaviate_retrieval import WeaviateRetrieval
from athina.steps.chroma_retrieval import ChromaRetrieval
from athina.steps.transform import ExtractJsonFromString, ExtractNumberFromString
from athina.steps.open_ai_assistant import OpenAiAssistant
from athina.steps.transcribe_speech_to_text import TranscribeSpeechToText
from athina.steps.search import Search
from athina.steps.code_execution import CodeExecution
from athina.steps.code_execution_v2 import CodeExecutionV2
from athina.steps.spider_crawl import SpiderCrawl

__all__ = [
    "Step",
    "Fn",
    "Debug",
    "Assert",
    "If",
    "Map",
    "Chain",
    "PromptExecution",
    "ExtractJsonFromString",
    "ExtractNumberFromString",
    "ApiCall",
    "ExtractEntities",
    "ClassifyText",
    "PineconeRetrieval",
    "QdrantRetrieval",
    "WeaviateRetrieval",
    "ChromaRetrieval",
    "OpenAiAssistant",
    "TranscribeSpeechToText",
    "Search",
    "CodeExecution",
    "CodeExecutionV2",
    "SpiderCrawl",
]



from typing import List, Optional
from athina.interfaces.athina import AthinaFilters
from athina.interfaces.data import DataPoint as BaseDataPoint
from .base_loader import BaseLoader
from dataclasses import asdict
from athina.services.athina_api_service import AthinaApiService
from llama_index.indices.query.base import BaseQueryEngine


class DataPoint(BaseDataPoint):
    """Data point for a single inference."""

    query: Optional[str]
    context: Optional[List[str]]
    response: Optional[str]
    expected_response: Optional[str]


class Loader(BaseLoader):
    """
    This class is a generic data loader for evals

    Attributes:
        col_query (str): The column name corresponding to the user's query.
        col_context (str): The column name corresponding to the retrieved context.
        col_response (str): The column name corresponding to the response.
        col_expected_response (str): The column name corresponding to the expected response.
        raw_dataset (dict): The raw dataset as loaded from the source.
        processed_dataset (list): The processed dataset with queries, context, response and other attributes if present.
    """

    def __init__(
        self,
        col_query="query",
        col_context="context",
        col_response="response",
        col_expected_response="expected_response",
    ):
        """
        Initializes the loader with specified or default column names.
        """
        self.col_query = col_query
        self.col_context = col_context
        self.col_response = col_response
        self.col_expected_response = col_expected_response
        self._raw_dataset = {}
        self._processed_dataset: List[DataPoint] = []

    def process(self) -> None:
        """
        Transforms the raw data into a structured format. Processes each entry from the raw dataset, and extracts attributes.
        """
        for raw_instance in self._raw_dataset:

            if self.col_query in raw_instance and not isinstance(
                raw_instance.get(self.col_query), str
            ):
                raise TypeError(f"'{self.col_query}' is not of type string.")
            if self.col_context in raw_instance:
                if not isinstance(raw_instance.get(self.col_context), list):
                    raise TypeError(f"'{self.col_context}' is not of type list.")
                if not all(
                    isinstance(element, str)
                    for element in raw_instance.get(self.col_context)
                ):
                    raise TypeError(
                        f"Not all elements in '{self.col_context}' are of type string."
                    )
            if self.col_response in raw_instance and not isinstance(
                raw_instance.get(self.col_response), str
            ):
                raise TypeError(f"'{self.col_response}' is not of type string.")
            if self.col_expected_response in raw_instance and not isinstance(
                raw_instance.get(self.col_expected_response), str
            ):
                raise TypeError(
                    f"'{self.col_expected_response}' is not of type string."
                )

            # Create a processed instance
            processed_instance = {
                "query": raw_instance.get(self.col_query, None),
                "context": raw_instance.get(self.col_context, None),
                "response": raw_instance.get(self.col_response, None),
                "expected_response": raw_instance.get(self.col_expected_response, None),
            }
            self._processed_dataset.append(processed_instance)

    def load_athina_inferences(
        self,
        filters: Optional[AthinaFilters] = None,
        limit: int = 10,
    ):
        """
        Load data from Athina API.
        By default, this will fetch the last 10 inferences from the API.
        """
        self._raw_dataset = AthinaApiService.fetch_inferences(
            filters=filters, limit=limit
        )
        for raw_dataset in self._raw_dataset:
            raw_dataset_dict = asdict(raw_dataset)

            context = (
                [str(raw_dataset_dict["context"])]
                if raw_dataset_dict["context"] is not None
                else None
            )
            processed_instance = {
                "query": raw_dataset_dict["user_query"],
                "context": context,
                "response": raw_dataset_dict["prompt_response"],
                "expected_response": raw_dataset_dict["expected_response"],
            }
            self._processed_dataset.append(processed_instance)
        return self._processed_dataset

    def _fetch_context_and_response_for_llama_index(
        self, query: str, query_engine: BaseQueryEngine
    ):
        """
        Fetches the context and response from the llama index query engine.
        """
        contexts = []
        query_engine_response = query_engine.query(query)
        response = query_engine_response.response
        for c in query_engine_response.source_nodes:
            text = c.node.get_content()
            contexts.append(text)

        return contexts, response

    def _generate_processed_instance_for_llama_index(
        self, raw_instance: dict, query_engine: BaseQueryEngine
    ) -> DataPoint:
        """
        Generates a processed instance for the llama index query engine.
        """
        if self.col_query not in raw_instance:
            raise ValueError(f"'{self.col_query}' not found in provided data.")
        if self.col_query in raw_instance and not isinstance(
            raw_instance.get(self.col_query), str
        ):
            raise TypeError(f"'{self.col_query}' is not of type string.")
        if self.col_expected_response in raw_instance and not isinstance(
            raw_instance.get(self.col_expected_response), str
        ):
            raise TypeError(f"'{self.col_expected_response}' is not of type string.")

        contexts, response = self._fetch_context_and_response_for_llama_index(
            raw_instance.get(self.col_query), query_engine
        )
        processed_instance = {
            "query": raw_instance.get(self.col_query),
            "context": contexts,
            "response": response,
            "expected_response": raw_instance.get(self.col_expected_response, None),
        }
        return processed_instance

    def load_from_llama_index(self, data: list, query_engine: BaseQueryEngine):
        """
        Load data from Llama Index.
        """
        if query_engine is None:
            raise ValueError("Query engine is not provided.")
        for raw_instance in data:
            processed_instance = self._generate_processed_instance_for_llama_index(
                raw_instance, query_engine
            )
            self._processed_dataset.append(processed_instance)
        return self._processed_dataset



import subprocess


def install():
    try:
        try:
            from guardrails import Guard
        except Exception as e:
            result = subprocess.call(["pip", "install", "-U", "guardrails-ai==0.4.2"])
            if result != 0:
                print("Guardrails installation failed")

        try:
            from guardrails.hub import GibberishText
        except Exception as e:
            result = subprocess.call(
                ["guardrails", "hub", "install", "hub://guardrails/gibberish_text"]
            )
            if result != 0:
                print(
                    "Guardrails installation failed. Ensure have the latest version of pip installed"
                )

        try:
            from guardrails.hub import SensitiveTopic
        except Exception as e:
            result = subprocess.call(
                ["guardrails", "hub", "install", "hub://guardrails/sensitive_topics"]
            )
            if result != 0:
                print(
                    "Guardrails installation failed. Ensure have the latest version of pip installed"
                )

        try:
            from guardrails.hub import NSFWText
        except Exception as e:
            result = subprocess.call(
                ["guardrails", "hub", "install", "hub://guardrails/nsfw_text"]
            )
            if result != 0:
                print(
                    "Guardrails installation failed. Ensure have the latest version of pip installed"
                )

        try:
            from guardrails.hub import ProfanityFree
        except Exception as e:
            result = subprocess.call(
                ["guardrails", "hub", "install", "hub://guardrails/profanity_free"]
            )
            if result != 0:
                print(
                    "Guardrails installation failed. Ensure have the latest version of pip installed"
                )
            else:
                print("Guardrails ProfanityFree validator installation successful")

        try:
            from guardrails.hub import DetectPII
        except Exception as e:
            result = subprocess.call(
                ["guardrails", "hub", "install", "hub://guardrails/detect_pii"]
            )
            if result != 0:
                print(
                    "Guardrails installation failed. Ensure have the latest version of pip installed"
                )
            else:
                print("Guardrails DetectPII validator installation successful")

        try:
            from guardrails.hub import ReadingTime
        except Exception as e:
            result = subprocess.call(
                ["guardrails", "hub", "install", "hub://guardrails/reading_time"]
            )
            if result != 0:
                print(
                    "Guardrails installation failed. Ensure have the latest version of pip installed"
                )
            else:
                print("Guardrails ReadingTime validator installation successful")

        try:
            from guardrails.hub import ToxicLanguage
        except Exception as e:
            result = subprocess.call(
                ["guardrails", "hub", "install", "hub://guardrails/toxic_language"]
            )
            if result != 0:
                print(
                    "Guardrails installation failed. Ensure have the latest version of pip installed"
                )
            else:
                print("Guardrails ToxicLanguage validator installation successful")

        try:
            from guardrails.hub import CorrectLanguage
        except Exception as e:
            result = subprocess.call(
                ["guardrails", "hub", "install", "hub://scb-10x/correct_language"]
            )
            if result != 0:
                print(
                    "Guardrails installation failed. Ensure have the latest version of pip installed"
                )
            else:
                print("Guardrails CorrectLanguage validator installation successful")

        try:
            from guardrails.hub import SecretsPresent
        except Exception as e:
            result = subprocess.call(
                ["guardrails", "hub", "install", "hub://guardrails/secrets_present"]
            )
            if result != 0:
                print(
                    "Guardrails installation failed. Ensure have the latest version of pip installed"
                )
            else:
                print("Guardrails SecretsPresent validator installation successful")

        try:
            from guardrails.hub import RestrictToTopic
        except Exception as e:
            result = subprocess.call(
                ["guardrails", "hub", "install", "hub://tryolabs/restricttotopic"]
            )
            if result != 0:
                print(
                    "Guardrails installation failed. Ensure have the latest version of pip installed"
                )
            else:
                print("Guardrails RestrictToTopic validator installation successful")

        try:
            from guardrails.hub import UnusualPrompt
        except Exception as e:
            result = subprocess.call(
                ["guardrails", "hub", "install", "hub://guardrails/unusual_prompt"]
            )
            if result != 0:
                print(
                    "Guardrails installation failed. Ensure have the latest version of pip installed"
                )
            else:
                print("Guardrails UnusualPrompt validator installation successful")

        try:
            from guardrails.hub import PolitenessCheck
        except Exception as e:
            result = subprocess.call(
                ["guardrails", "hub", "install", "hub://guardrails/politeness_check"]
            )
            if result != 0:
                print(
                    "Guardrails installation failed. Ensure have the latest version of pip installed"
                )
            else:
                print("Guardrails PolitenessCheck validator installation successful")

        try:
            import nltk
            import ssl

            try:
                _create_unverified_https_context = ssl._create_unverified_context
            except AttributeError:
                pass
            else:
                ssl._create_default_https_context = _create_unverified_https_context
            nltk.download("punkt")
        except Exception as e:
            print(f"Error: {e}")
    except Exception as e:
        print(f"Error: {e}")



import os
from pydantic import BaseModel
from typing import List, Dict, Any, Optional, Union
from athina.helpers.json import JsonExtractor
from athina.interfaces.model import Model
from athina.steps.base import Step
from athina.llms.abstract_llm_service import AbstractLlmService
from athina.keys import OpenAiApiKey
from athina.llms.openai_service import OpenAiService
from jinja2 import Environment
from athina.helpers.jinja_helper import PreserveUndefined
from athina.steps.transform import ExtractJsonFromString, ExtractNumberFromString
import traceback
import json


class TextContent(BaseModel):
    type: str
    text: str


class ImageContent(BaseModel):
    type: str = "image"
    image_url: Union[str, Dict[str, str]]

    def to_api_format(self):
        if isinstance(self.image_url, dict):
            return {"type": "image_url", "image_url": self.image_url}
        return {"type": "image_url", "image_url": {"url": self.image_url}}


Content = Union[str, List[Union[TextContent, ImageContent]]]


class PromptMessage(BaseModel):
    role: str
    content: Optional[Content] = None
    tool_call: Optional[str] = None

    def to_api_format(self) -> dict:
        """Convert the message to the format expected by the OpenAI API"""
        if self.content is None:
            return {"role": self.role}

        if isinstance(self.content, str):
            return {"role": self.role, "content": self.content}

        if isinstance(self.content, list):
            formatted_content = []
            for item in self.content:
                if isinstance(item, TextContent):
                    formatted_content.append({"type": "text", "text": item.text})
                elif isinstance(item, ImageContent):
                    formatted_content.append(item.to_api_format())
            return {"role": self.role, "content": formatted_content}


class ModelOptions(BaseModel):
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    frequency_penalty: Optional[float] = None
    presence_penalty: Optional[float] = None


class ToolConfig(BaseModel):
    tool_choice: Optional[Union[str, Dict[str, Any]]] = None
    tools: Optional[List[Any]] = None


class PromptTemplate(BaseModel):
    messages: List[PromptMessage]
    env: Environment = None

    class Config:
        arbitrary_types_allowed = True

    @staticmethod
    def simple(message: str) -> "PromptTemplate":
        """Create a PromptTemplate from a string representation."""
        messages = [PromptMessage(role="user", content=message)]
        return PromptTemplate(messages=messages)

    def resolve(self, **kwargs) -> List[PromptMessage]:
        """Render the template with given variables."""

        # Create a custom Jinja2 environment with double curly brace delimiters and PreserveUndefined
        self.env = Environment(
            variable_start_string="{{",
            variable_end_string="}}",
            undefined=PreserveUndefined,
        )

        final_messages = []
        for message in self.messages:
            if message.role == "import":
                # Find the value wrapped in {{}}
                import_key = message.content.strip("{}")

                # Find the value in the row
                if import_key in kwargs:
                    value = kwargs[import_key]

                    # Check if it is a list/array
                    if isinstance(value, list):
                        # Iterate over the list and create a new PromptMessage for each item
                        for item in value:
                            if isinstance(item, dict):
                                # If item has tool_call, then parse tool_call and create a new PromptMessage
                                if "tool_call" in item:
                                    try:
                                        tool_call_message = PromptMessage(
                                            role=item["role"],
                                            tool_call=self.env.from_string(
                                                item.get("tool_call")
                                            ).render(**kwargs),
                                        )
                                        final_messages.append(tool_call_message)
                                    except Exception as e:
                                        print(f"Error parsing tool_call: {e}")
                                else:
                                    new_message = PromptMessage(**item)
                                    final_messages.append(new_message)
            else:
                final_messages.append(message)

        resolved_messages = []
        for message in final_messages:
            if message.content is None:
                resolved_messages.append(message)
            elif isinstance(message.content, str):
                content_template = self.env.from_string(message.content)
                content = content_template.render(**kwargs)
                resolved_message = PromptMessage(role=message.role, content=content)
                resolved_messages.append(resolved_message)
            elif isinstance(message.content, list):
                resolved_content = []
                for item in message.content:
                    if isinstance(item, TextContent):
                        content_template = self.env.from_string(item.text)
                        resolved_text = content_template.render(**kwargs)
                        resolved_content.append(
                            TextContent(text=resolved_text, type="text")
                        )
                    elif isinstance(item, ImageContent):
                        resolved_content.append(item)
                resolved_message = PromptMessage(
                    role=message.role, content=resolved_content
                )
                resolved_messages.append(resolved_message)

        return resolved_messages


class PromptExecution(Step):
    """
    Step that executes a prompt using an LLM service.

    Attributes:
        llm_service (AbstractLlmService): The LLM service to use for prompt execution.
        template (PromptTemplate): The template to render the prompt.
        model (str): The model to use for the LLM service.
    """

    llm_service: AbstractLlmService = None
    template: PromptTemplate
    model: str
    model_options: ModelOptions
    tool_config: Optional[ToolConfig] = None
    response_format: Optional[dict] = None
    name: Optional[str] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if kwargs.get("llm_service"):
            self.llm_service = kwargs.get("llm_service")
        else:
            self.llm_service = OpenAiService()

    class Config:
        arbitrary_types_allowed = True

    @staticmethod
    def simple(message: str, model: str = Model.GPT4_O.value) -> "PromptExecution":
        OpenAiApiKey.set_key(os.getenv("OPENAI_API_KEY"))
        openai_service = OpenAiService()
        return PromptExecution(
            llm_service=openai_service,
            template=PromptTemplate.simple(message),
            model=model,
        )

    def execute(self, input_data: dict, **kwargs) -> str:
        """Execute a prompt with the LLM service."""
        if input_data is None:
            input_data = {}

        if not isinstance(input_data, dict) and self.input_key:
            raise ValueError("PromptExecution Error: Input data must be a dictionary")

        try:
            messages = self.template.resolve(**input_data)
            # Convert messages to API format
            api_formatted_messages = [msg.to_api_format() for msg in messages]

            llm_service_response = self.llm_service.chat_completion(
                messages,
                model=self.model,
                **self.model_options.model_dump(),
                **(self.tool_config.model_dump() if self.tool_config else {}),
                **({"response_format": self.response_format}),
                **(
                    kwargs.get("search_domain_filter", {})
                    if isinstance(kwargs.get("search_domain_filter"), dict)
                    else {}
                ),
            )
            llmresponse = llm_service_response["value"]
            output_type = kwargs.get("output_type", None)
            error = None
            if output_type:
                if output_type == "string":
                    if not isinstance(llmresponse, str):
                        error = "LLM response is not a string"
                    response = llmresponse

                elif output_type == "number":
                    extracted_response = ExtractNumberFromString().execute(llmresponse)
                    if not isinstance(extracted_response, (int, float)):
                        error = "LLM response is not a number"
                    response = extracted_response

                elif output_type == "array":
                    extracted_response = ExtractJsonFromString().execute(llmresponse)
                    if not isinstance(extracted_response, list):
                        error = "LLM response is not an array"
                    response = extracted_response

                elif output_type == "object":
                    extracted_response = ExtractJsonFromString().execute(llmresponse)
                    if not isinstance(extracted_response, dict):
                        error = "LLM response is not an object"
                    response = extracted_response

            elif not isinstance(llmresponse, str):
                error = "LLM service response is not a string"

            else:
                response = llmresponse

            if error:
                return {"status": "error", "data": error}
            else:
                return {
                    "status": "success",
                    "data": response,
                    "metadata": json.loads(llm_service_response.get("metadata", {})),
                }
        except Exception as e:
            traceback.print_exc()
            return {"status": "error", "data": str(e)}



# Step to classify text into one of the provided labels.
from typing import Union, Dict, Any
from athina.steps import Step
import marvin


class ClassifyText(Step):
    """
    Step that classifies text into one of the labels provided to the step.

    Attributes:
        input_column: The row's column to classify.
        labels: The labels to classify the text into.
        llm_api_key: The API key for the language model.
        language_model_id: The language model ID to use for classification.
    """

    labels: list[str]
    input_column: str
    llm_api_key: str
    language_model_id: str

    def execute(self, input_data: Any) -> Union[Dict[str, Any], None]:
        """Classify the text and return the label."""

        if input_data is None:
            input_data = {}

        if not isinstance(input_data, dict):
            raise TypeError("Input data must be a dictionary.")

        input_text = input_data.get(self.input_column, None)

        if input_text is None:
            return None

        marvin.settings.openai.api_key = self.llm_api_key
        marvin.settings.openai.chat.completions.model = self.language_model_id

        try:
            result = marvin.classify(
                input_text,
                labels=self.labels,
            )
            return {
                "status": "success",
                "data": result,
            }
        except Exception as e:
            return {
                "status": "error",
                "data": str(e),
            }



import re
import math
from abc import ABC, abstractmethod


class Comparator(ABC):
    @abstractmethod
    def compare(self, string1, string2):
        pass


class CosineSimilarity(Comparator):
    def compare(self, string1, string2):
        # Tokenize and create a combined set of unique words
        combined_set = self._create_combined_set(string1, string2)
        # Vectorize the strings
        vector1 = self._vectorize(string1, combined_set)
        vector2 = self._vectorize(string2, combined_set)
        dot_product = sum(p * q for p, q in zip(vector1, vector2))
        magnitude_vec1 = math.sqrt(sum([val**2 for val in vector1]))
        magnitude_vec2 = math.sqrt(sum([val**2 for val in vector2]))
        if magnitude_vec1 * magnitude_vec2 == 0:
            # Avoid division by zero
            return 0
        return dot_product / (magnitude_vec1 * magnitude_vec2)

    def _tokenize(self, string):
        """
        Tokenize the input string into a list of words.

        Args:
            string (str): The string to tokenize.

        Returns:
            list: A list of lowercased words from the string.
        """
        return re.findall(r"\b\w+\b", string.lower())

    def _create_combined_set(self, string1, string2):
        return set(self._tokenize(string1)).union(set(self._tokenize(string2)))

    def _vectorize(self, string, combined_set):
        tokenized = self._tokenize(string)
        vector = [tokenized.count(word) for word in combined_set]
        return vector


class NormalisedLevenshteinSimilarity(Comparator):
    def compare(self, string1, string2):
        return 1 - self._normalised_levenshtein_distance(string1, string2)

    def _normalised_levenshtein_distance(self, str1, str2):
        m, n = len(str1), len(str2)
        # Create a matrix to store the distances
        dp = [[0] * (n + 1) for _ in range(m + 1)]
        # Initialize the first row and first column
        for i in range(m + 1):
            dp[i][0] = i
        for j in range(n + 1):
            dp[0][j] = j
        # Calculate the distance
        for i in range(1, m + 1):
            for j in range(1, n + 1):
                if str1[i - 1] == str2[j - 1]:
                    dp[i][j] = dp[i - 1][j - 1]
                else:
                    dp[i][j] = 1 + min(dp[i - 1][j], dp[i][j - 1], dp[i - 1][j - 1])
        if len(str1) >= len(str2):
            return dp[m][n] / len(str1)
        else:
            return dp[m][n] / len(str2)


class JaroWincklerSimilarity(Comparator):
    def compare(self, string1, string2):
        return self._jaro_winckler_similarity(string1, string2)

    def _jaro_winckler_similarity(self, str1, str2):
        len1 = len(str1)
        len2 = len(str2)
        if len1 == 0 or len2 == 0:
            return 0.0
        max_dist = (max(len(str1), len(str2)) // 2) - 1
        match = 0
        hash_str1 = [0] * len(str1)
        hash_str2 = [0] * len(str2)
        for i in range(len1):
            for j in range(max(0, i - max_dist), min(len2, i + max_dist + 1)):
                if str1[i] == str2[j] and hash_str2[j] == 0:
                    hash_str1[i] = 1
                    hash_str2[j] = 1
                    match += 1
                    break
        if match == 0:
            return 0.0
        t = 0
        point = 0
        for i in range(len1):
            if hash_str1[i]:
                while hash_str2[point] == 0:
                    point += 1
                if str1[i] != str2[point]:
                    point += 1
                    t += 1
        t //= 2
        return (match / len1 + match / len2 + (match - t) / match) / 3.0


class JaccardSimilarity(Comparator):
    def compare(self, string1, string2):
        return self._jaccard_similarity(string1, string2)

    def _jaccard_similarity(self, str1, str2):
        str1_tokens = set(str1.split())
        str2_tokens = set(str2.split())
        return len(str1_tokens.intersection(str2_tokens)) / len(
            str1_tokens.union(str2_tokens)
        )


class SorensenDiceSimilarity(Comparator):
    def compare(self, string1, string2):
        return self._sorensen_dice_similarity(string1, string2)

    def _sorensen_dice_similarity(self, str1, str2):
        str1_tokens = set(str1.split())
        str2_tokens = set(str2.split())
        return (
            2
            * len(str1_tokens.intersection(str2_tokens))
            / (len(str1_tokens) + len(str2_tokens))
        )



from athina.interfaces.model import Model
from ..ragas_evaluator import RagasEvaluator
from athina.evals.eval_type import RagasEvalTypeId
from athina.metrics.metric_type import MetricType
from ragas.metrics import context_relevancy
from typing import List, Optional

"""
RAGAS Context Relevancy Docs: https://docs.ragas.io/en/latest/concepts/metrics/context_relevancy.html
RAGAS Context Relevancy Github: https://github.com/explodinggradients/ragas/blob/main/src/ragas/metrics/_context_relevancy.py
"""


class RagasContextRelevancy(RagasEvaluator):
    """
    This evaluator calculates the relevancy of the context with respect to the user query.
    """

    @property
    def name(self):
        return RagasEvalTypeId.RAGAS_CONTEXT_RELEVANCY.value

    @property
    def display_name(self):
        return "Ragas Context Relevancy"

    @property
    def metric_ids(self) -> List[str]:
        return [MetricType.RAGAS_CONTEXT_RELEVANCY.value]

    @property
    def ragas_metric(self):
        return context_relevancy

    @property
    def ragas_metric_name(self):
        return "context_relevancy"

    @property
    def default_model(self):
        return Model.GPT35_TURBO.value

    @property
    def required_args(self):
        return ["query", "context"]

    @property
    def examples(self):
        return None

    @property
    def grade_reason(self) -> str:
        return "This metric is calulated by dividing the number of sentences in context that are relevant for answering the given query by the total number of sentences in the retrieved context"

    def is_failure(self, score) -> Optional[bool]:
        return (
            bool(score < self._failure_threshold)
            if self._failure_threshold is not None
            else None
        )

    def generate_data_to_evaluate(self, context, query, **kwargs) -> dict:
        """
        Generates data for evaluation.

        :param context: list of strings of retrieved context
        :param query: user query
        :return: A dictionary with formatted data for evaluation
        """
        data = {"contexts": [context], "question": [query]}
        return data



from openai import OpenAI
from retrying import retry
from timeout_decorator import timeout
from athina.helpers.json import JsonHelper
from athina.keys import OpenAiApiKey
from athina.interfaces.model import Model
from athina.errors.exceptions import NoOpenAiApiKeyException
from .abstract_llm_service import AbstractLlmService
import json
import time
from litellm import cost_per_token

DEFAULT_TEMPERATURE = 0.0


class OpenAiService(AbstractLlmService):
    _instance = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(OpenAiService, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        openai_api_key = OpenAiApiKey.get_key()
        if openai_api_key is None:
            raise NoOpenAiApiKeyException()
        self.openai = OpenAI(api_key=openai_api_key)

    def embeddings(self, text: str, model: str) -> list:
        """
        Fetches response from OpenAI's Embeddings API.
        """
        try:
            response = self.openai.embeddings.create(
                model=model, input=text, encoding_format="float"
            )
            return response.data[0].embedding
        except Exception as e:
            print(f"Error in Embeddings: {e}")
            raise e

    def _process_response(self, response, start_time, model):
        end_time = time.time()
        completion_time = (end_time - start_time) * 1000
        prompt_tokens_cost_usd_dollar, completion_tokens_cost_usd_dollar = (
            cost_per_token(
                model=model,
                prompt_tokens=response.usage.prompt_tokens,
                completion_tokens=response.usage.completion_tokens,
            )
        )
        metadata = json.dumps(
            {
                "usage": {
                    "completion_tokens": response.usage.completion_tokens,
                    "prompt_tokens": response.usage.prompt_tokens,
                    "total_tokens": response.usage.total_tokens,
                },
                "cost": {
                    "prompt_tokens_cost_usd_dollar": prompt_tokens_cost_usd_dollar,
                    "completion_tokens_cost_usd_dollar": completion_tokens_cost_usd_dollar,
                    "total_cost_usd_dollar": prompt_tokens_cost_usd_dollar
                    + completion_tokens_cost_usd_dollar,
                },
                "response_time": completion_time,
            }
        )
        if response.choices[0].finish_reason == "tool_calls":
            tool_calls = [
                call.model_dump() for call in response.choices[0].message.tool_calls
            ]
            tool_calls_data = [
                {
                    "arguments": call["function"]["arguments"],
                    "name": call["function"]["name"],
                }
                for call in tool_calls
            ]
            return {"value": json.dumps(tool_calls_data), "metadata": metadata}
        else:
            prompt_response = response.choices[0].message.content
            if not prompt_response:
                if response.choices[0].message.tool_calls:
                    tool_calls = [
                        call.model_dump()
                        for call in response.choices[0].message.tool_calls
                    ]
                    tool_calls_data = [
                        {
                            "arguments": call["function"]["arguments"],
                            "name": call["function"]["name"],
                        }
                        for call in tool_calls
                    ]
                    return {"value": json.dumps(tool_calls_data), "metadata": metadata}
                else:
                    return {
                        "value": json.dumps(response.choices[0].message.__dict__),
                        "metadata": metadata,
                    }
            return {"value": prompt_response, "metadata": metadata}

    @retry(stop_max_attempt_number=3, wait_fixed=2000)
    def chat_completion(self, messages, model, **kwargs) -> str:
        """
        Fetches response from OpenAI's ChatCompletion API.
        """
        if "temperature" not in kwargs:
            kwargs["temperature"] = DEFAULT_TEMPERATURE
        try:
            start_time = time.time()
            response = self.openai.chat.completions.create(
                model=model, messages=messages, **kwargs
            )
            return self._process_response(response, start_time, model)
        except Exception as e:
            print(f"Error in ChatCompletion: {e}")
            raise e

    @retry(stop_max_attempt_number=3, wait_fixed=2000)
    def chat_completion_json(self, messages, model, **kwargs) -> str:
        """
        Fetches response from OpenAI's ChatCompletion API using JSON mode.
        """
        if "temperature" not in kwargs:
            kwargs["temperature"] = DEFAULT_TEMPERATURE
        try:
            start_time = time.time()
            response = self.openai.chat.completions.create(
                model=model,
                messages=messages,
                response_format={"type": "json_object"},
                **kwargs,
            )
            return self._process_response(response, start_time, model)
        except Exception as e:
            print(f"Error in JSON ChatCompletion: {e}")
            raise e

    def json_completion(self, messages, model, **kwargs):
        """
        Fetches response from OpenAI's ChatCompletion API using JSON mode.
        """
        if "temperature" not in kwargs:
            kwargs["temperature"] = DEFAULT_TEMPERATURE
        try:
            if Model.supports_json_mode(model):
                chat_completion_result = self.chat_completion_json(
                    model=model,
                    messages=messages,
                    **kwargs,
                )
            else:
                chat_completion_result = self.chat_completion(
                    model=model,
                    messages=messages,
                    **kwargs,
                )
            chat_completion_response = chat_completion_result["value"]
            # Extract JSON object from LLM response
            eval_response = JsonHelper.extract_json_from_text(chat_completion_response)
            if "metadata" in chat_completion_result:
                metadata = json.loads(chat_completion_result["metadata"])
                eval_response["metadata"] = metadata
            return eval_response

        except Exception as e:
            print(f"Error in ChatCompletion: {e}")
            raise e



# Step to make an external api call
import json
import time
from typing import Union, Dict, List, Any, Iterable, Optional
import requests
from athina.steps import Step
from jinja2 import Environment
from athina.helpers.jinja_helper import PreserveUndefined


def prepare_input_data(data):
    return {key: json.dumps(value) if isinstance(value, (list, dict)) else value
        for key, value in data.items()}


class Search(Step):
    """
    Step that makes a search API Call to https://exa.ai/.

    Attributes:
        query: The query string.
        type: The Type of search, 'keyword', 'neural', or 'auto' (decides between keyword and neural). Default neural.
        category: Optional data category to focus on, with higher comprehensivity and data cleanliness. Categories right now include company, research paper, news article, linkedin profile, github, tweet, movie, song, personal site, pdf and financial report.
        numResults: Optional number of search results to return. Default 10. Max 10 for basic plans. Up to thousands for custom plans.
        excludedDomains: Optional list of domains to exclude in the search. If specified, results will not include any from these domains.
        includedDomains: Optional list of domains to include in the search. If specified, results will only come from these domains..
        excludeText: Optional list of strings that must not be present in webpage text of results. Currently, only 1 string is supported, of up to 5 words.
        includeText: Optional list of strings that must be present in webpage text of results. Currently, only 1 string is supported, of up to 5 words.
        x_api_key: The API key to use for the request.
    """

    query: str
    type: Optional[str] = "neural"
    category: Optional[str] = None
    numResults: Optional[int] = 10
    excludedDomains: Optional[List[str]] = None
    includedDomains: Optional[List[str]] = None
    excludeText: Optional[List[str]] = None
    includeText: Optional[List[str]] = None
    x_api_key: str
    env: Environment = None

    class Config:
        arbitrary_types_allowed = True
       

    def execute(self, input_data: Any) -> Union[Dict[str, Any], None]:
        """Make an Search API call and return the response."""

        if input_data is None:
            input_data = {}

        if not isinstance(input_data, dict):
            raise TypeError("Input data must be a dictionary.")

        # Create a custom Jinja2 environment with double curly brace delimiters and PreserveUndefined
        self.env = Environment(
            variable_start_string='{{', 
            variable_end_string='}}',
            undefined=PreserveUndefined
        )

        body={
            "query": self.query,
            "type": self.type,
            "category": self.category,
            "numResults": self.numResults,
            "excludedDomains": self.excludedDomains,
            "includedDomains": self.includedDomains,
            "excludeText": self.excludeText,
            "includeText": self.includeText
        }
        prepared_body = None
        # Add a filter to the Jinja2 environment to convert the input data to JSON
        body_template = self.env.from_string(json.dumps(body))
        prepared_input_data = prepare_input_data(input_data)
        prepared_body = body_template.render(**prepared_input_data)
        

        retries = 2  # number of retries
        timeout = 30  # seconds
        for attempt in range(retries):
            try:
                response = requests.post(
                    url= "https://api.exa.ai/search",
                    headers={
                        "Content-Type": "application/json",
                        "x-api-key": self.x_api_key
                    },
                    json=json.loads(prepared_body, strict=False) if prepared_body else None,
                    timeout=timeout,
                )
                if response.status_code >= 400:
                    # If the status code is an error, return the error message
                    return {
                        "status": "error",
                        "data": f"Failed to make the API call.\nStatus code: {response.status_code}\nError:\n{response.text}",
                    }
                try:
                    json_response = response.json()
                    # If the response is JSON, return the JSON data
                    return {
                        "status": "success",
                        "data": json_response,
                    }
                except json.JSONDecodeError:
                    # If the response is not JSON, return the text
                    return {
                        "status": "success",
                        "data": response.text,
                    }
            except requests.Timeout:
                if attempt < retries - 1:
                    time.sleep(2)
                    continue
                # If the request times out after multiple attempts, return an error message
                return {
                    "status": "error",
                    "data": "Failed to make the API call.\nRequest timed out after multiple attempts.",
                }
            except Exception as e:
                # If an exception occurs, return the error message
                return {
                    "status": "error",
                    "data": f"Failed to make the API call.\nError: {e.__class__.__name__}\nDetails:\n{str(e)}",
                }
